#!/usr/bin/python
# Copyright (C) 2001,  Christof Meerwald <cmeerw@web.de>
# http://xpcom2corba.sunsite.dk/

# XPCOM-CORBA is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 dated June, 1991.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.

import os, string, sys

from omniidl import idlast, idlvisitor, idlutil, idltype, output
from corba_tmpl import *
import xpcorba_type
convert_scoped2flat = xpcorba_type.convert_scoped2flat

keywords = ['and', 'and_eq', 'asm', 'auto', 'bitand', 'bitor', 'bool',\
            'break', 'case', 'catch', 'char', 'class', 'compl', 'const',\
            'const_cast', 'continue', 'default', 'delete', 'do', 'double',\
            'dynamic_cast', 'else', 'enum', 'explicit', 'export', 'extern',\
            'false', 'float', 'for', 'friend', 'goto', 'if', 'inline', 'int',\
            'long', 'mutable', 'namespace', 'new', 'not', 'not_eq',\
            'operator', 'or', 'or_eq', 'private', 'protected', 'public',\
            'register', 'reinterpret_cast', 'return', 'short', 'signed',\
            'sizeof', 'static', 'static_cast', 'struct', 'switch',\
            'template', 'this', 'throw', 'true', 'try', 'typedef', 'typeid',\
            'typename', 'union', 'unsigned', 'using', 'virtual', 'void',\
            'volatile', 'wchar_t', 'while', 'xor', 'xor_eq']
keyword_map = {}
for kw in keywords:
    keyword_map[kw] = '_cxx_' + kw

REPOID_ISUPPORTS = 'DCE:00000000-0000-0000-c000-000000000046:1'


class CORBA_Proxy_Visitor(idlvisitor.AstVisitor):
    def __init__(self):
        self._header = None
        self._impl = None
        self._toplevel = []
        self._included = ['xpcom.idl']

    def makeCxxName(self, name):
        if not keyword_map.has_key(name):
            return name
        return keyword_map[name]

    def visitAST(self, node):
        directory, basename = os.path.split(node.file())
        if string.lower(basename[-4:]) == '.idl':
            basename = basename[:-4]

        self._impl = output.Stream(open(basename + '_corba.cc', 'w'), 2)
        self._header = output.Stream(open(basename + '_corba.h', 'w'), 2)
        gbasename = '';
        for i in basename:
            if (i >= 'A' and i <= 'Z') or (i >= 'a' and i <= 'z'):
                gbasename = gbasename + i
        guardname = 'HEADER__' + string.upper(gbasename) + '_CORBA__INCLUDED'

        self._header.out(TMPL_proxy_h_file_header, basename=basename,
                         guardname=guardname,
                         CORBAprefix='', CORBAsuffix='.hh',
                         XPCOMprefix='I', XPCOMsuffix='.h')

        self._impl.out(TMPL_proxy_c_file_header, basename=basename)

        for n in node.declarations():
            if n.mainFile():
                n.accept(self)
            else:
                filename = n.file()
                pos = string.rfind(filename, '/')
                if pos != -1:
                    filename = filename[pos + 1:]

                if not filename in self._included:
                    self._included.append(filename)

                    if filename[-4:] == '.idl':
                        filename = filename[:-4]

                    self._header.out('#include "@filename@_corba.h"\n#include "@filename@_xpcom.h"\n',
                                   filename=filename)

        for s, args in self._toplevel:
            apply(self._header.out, [s], args)
        self._toplevel = []

        self._header.out(TMPL_proxy_h_file_footer)

        self._impl = None
        self._header = None


    def visitModule(self, node):
        for n in node.definitions():
            if n.mainFile():
                n.accept(self)


    def visitTypedef(self, node):
        for decl in node.declarators():
            if isinstance(node.aliasType(), idltype.Sequence):
                childtype = node.aliasType().seqType().unalias()

                corba_name = idlutil.ccolonName(decl.scopedName())
                xpchild_type = xpcorba_type.get_type(childtype)
                xpcom_childname = xpchild_type.type_xpcom()

                if isinstance(childtype, idltype.Declared):
                    corba_childname = xpchild_type._corbaname
                    convert_child = xpchild_type.copy_corba2xpcom('corba_obj[i]', '_xpcom_child')
                    self._toplevel.append((TMPL_proxy_h_sequence_objref,
                                           {'corba_name': corba_name,
                                            'xpcom_childname': xpcom_childname,
                                            'corba_childname': corba_childname,
                                            'convert_child': convert_child}))
                else:
                    self._toplevel.append((TMPL_proxy_h_sequence,
                                           {'corba_name': corba_name,
                                            'xpcom_childname': xpcom_childname}))

    def visitStruct(self, node):
        corba_name = idlutil.ccolonName(node.scopedName())
        xpcom_name = convert_scoped2flat(node.scopedName())


        self._header.out(TMPL_proxy_h_struct_proxy,
                         corba_name=corba_name,
                         intfname=xpcom_name,
                         uppername=string.upper(xpcom_name))
        self._header.inc_indent()
        self._impl.out(TMPL_proxy_c_struct_header,
                       corba_name=corba_name,
                       intfname=xpcom_name)

        for member in node.members():
            identifier = member.declarators()[0].identifier()
            xpcom_type = xpcorba_type.get_type(member.memberType())

            self._impl.out('NS_IMETHODIMP @intfname@__XPCOM_Proxy::Get@attrname@(@param@)',
                           intfname=xpcom_name,
                           attrname=string.upper(identifier[0])+identifier[1:],
                           param=xpcom_type.decl_xpcom('_retval', 0, 1))
            self._impl.out('{')
            self._impl.inc_indent()
            if isinstance(xpcom_type, xpcorba_type.Sequence) or isinstance(xpcom_type, xpcorba_type.WString) or isinstance(xpcom_type, xpcorba_type.String):
                self._impl.out(xpcom_type.copy_corba2xpcom('value_.' + identifier,
                                                           '_retval'))
            else:
                self._impl.out(xpcom_type.copy_corba2xpcom('value_.' + identifier,
                                                           '*_retval'))
            self._impl.out('return NS_OK;')
            self._impl.dec_indent()
            self._impl.out('}\n')

        self._header.dec_indent()


        self._header.out(TMPL_proxy_h_struct_c2x_header,
                         corba_name=corba_name, xpcom_name=xpcom_name)
        self._header.inc_indent()
        self._header.out('*xpcom_obj = new @xpcom_name@__XPCOM_Proxy(corba_obj); NS_ADDREF(*xpcom_obj);',
                         xpcom_name=xpcom_name)
        self._header.dec_indent()
        self._header.out(TMPL_proxy_h_struct_c2x_footer)


    def visitInterface(self, node):
        # check inheritance tree: every interface must somehow inherit
        # from nsISupports (XPCOM::IObject)
        if node.inherits():
            to_check = node.inherits()
            while to_check:
                if to_check[0].repoId() == REPOID_ISUPPORTS:
                    to_check = to_check[1:]
                else:
                    if not to_check[0].inherits():
                        return
                    else:
                        to_check = to_check[1:] + to_check[0].inherits()
        else:
            return

        # We have a list of supported interfaces. This list consists of
        # every ancestor (including the interface itself).
        supportedInterfaces = [node]
        interfaceStack = node.inherits()[:]
        while len(interfaceStack) != 0:
            i = interfaceStack.pop()
            # Ignore the ISupports interface...
            if i.repoId() == REPOID_ISUPPORTS:
                continue
            # Never duplicate interfaces...
            if i in supportedInterfaces:
                continue
            supportedInterfaces.append(i)
            interfaceStack = interfaceStack + i.inherits()

        directInterfaces = [node];
        ni = node.inherits()
        while len(ni) > 0:
            n = ni[0]
            if n.repoId() == REPOID_ISUPPORTS:
                break
            directInterfaces.append(n)
            ni = n.inherits()

        name = node.identifier()
        fqname = idlutil.ccolonName(node.scopedName())
        self.fqname = fqname
        if len(node.scopedName()) > 1:
            namespace_begin = 'namespace CORBA_Proxy__' +\
                              self.makeCxxName(node.scopedName()[0]) + '\n{\n'
            namespace_end = '}\n'
            for scope in node.scopedName()[1:-1]:
                namespace_begin = namespace_begin + 'namespace ' +\
                                  self.makeCxxName(scope) + '\n{\n'
                namespace_end = namespace_end + '}\n'
            proxyname = name
        else:
            namespace_begin = ''
            namespace_end = ''
            proxyname = 'CORBA_Proxy__' + name
        xpcomintfname = self.makeCxxName(convert_scoped2flat(node.scopedName()))
        
        self._header.out(TMPL_proxy_h_intf_header,
                         proxyname=proxyname, fqname=fqname,
                         namespace_begin=namespace_begin,
                         namespace_end=namespace_end,
                         xpcomintf=xpcomintfname)

        self._impl.out(TMPL_proxy_c_intf,
                       proxyname=proxyname, fqname=fqname,
                       namespace_begin=namespace_begin,
                       namespace_end=namespace_end,
                       xpcomintf=xpcomintfname, repoid=node.repoId())

        self._impl.out(TMPL_proxy_c_intf_opnames_header,
                       fqname=fqname)
        self._impl.inc_indent()

        # We include all ancestors...
        for i in supportedInterfaces:
            for c in i.callables():
                if isinstance(c, idlast.Operation):
                    self._impl.out('"' + c.identifier() + '",')
                elif isinstance(c, idlast.Attribute):
                    for identifier in c.identifiers():
                        self._impl.out('"_get_' + identifier + '",')
                        if not c.readonly():
                            self._impl.out('"_set_' + identifier + '",')

        self._impl.dec_indent()
        self._impl.out(TMPL_proxy_c_intf_opnames_footer)

        self._impl.out(TMPL_proxy_c_intf_baserepoids_header,
                       fqname=fqname)
        self._impl.inc_indent()

        for intf in node.inherits():
            self._impl.out('"' + intf.repoId() + '",')

        self._impl.dec_indent()
        self._impl.out(TMPL_proxy_c_intf_baserepoids_footer)


        self._header.inc_indent()

        for d in node.declarations():
            d.accept(self)

        for i in supportedInterfaces:
            self.need_qi = not (i in directInterfaces)
            if self.need_qi:
                xpcomintf = self.makeCxxName(convert_scoped2flat(i.scopedName()))
                self.qiline = 'nsCOMPtr<' + xpcomintf + '> xpcom_obj_qi_(' +\
                              'do_QueryInterface(xpcom_obj_));'
            for c in i.callables():
                c.accept(self)

        self._header.dec_indent()
        self._header.out(TMPL_proxy_h_intf_footer,
                         xpcomintf=xpcomintfname, fqname=fqname,
                         namespace_begin=namespace_begin,
                         namespace_end=namespace_end)

    def getAppropriateObject(self, node):
        if self.need_qi:
            self._impl.out(self.qiline)
            return 'xpcom_obj_qi_'
        else:
            return 'xpcom_obj_'

    def visitAttribute(self, node):
        corba_type = xpcorba_type.get_type(node.attrType())

        retspec = corba_type.decl_corba('', 0, 1)
        callparams = corba_type.call_xpcom_param('_retval', 1, 1)

        xpcom_preinvoke = corba_type.xpcom_preinvoke('_retval', 0, 1, 0)
        xpcom_postinvoke = corba_type.xpcom_postinvoke('_retval', 1, 1, 0)

        for name in node.identifiers():
            capname = 'Get' + string.upper(name[0]) + name[1:]
            fqname = idlutil.ccolonName(node.declarators()[0].scopedName()[:-1])

            self._header.out(TMPL_proxy_h_operation, name=self.makeCxxName(name),
                             capname=capname,
                             paramspec='', retspec=retspec)

            self._impl.out(TMPL_proxy_c_operation_header,
                           name=self.makeCxxName(name),
                           fqname=self.makeCxxName(self.fqname),
                           capname=capname, paramspec='', retspec=retspec)
            self._impl.inc_indent()

            self._impl.out(corba_type.type_corba() + ' _retval;')

            self._impl.out(xpcom_preinvoke)
            objname = self.getAppropriateObject(node)
            self._impl.out('_result = ' + objname + '->@capname@(@paramlist@);',
                           capname=capname, paramlist=callparams)

            self._impl.out(xpcom_postinvoke)
            self._impl.out('return _retval;')

            self._impl.dec_indent()
            self._impl.out(TMPL_proxy_c_operation_footer)


        if not node.readonly():
            paramspec = corba_type.decl_corba('_newval', 1, 0)
            callparams = corba_type.call_xpcom_param('_newval', 1, 0)

            xpcom_preinvoke = corba_type.xpcom_preinvoke('_newval', 1, 0, 0)
            xpcom_postinvoke = corba_type.xpcom_postinvoke('_newval', 1, 0, 0)

            for name in node.identifiers():
                capname = 'Set' + string.upper(name[0]) + name[1:]
                fqname = self.makeCxxName(idlutil.ccolonName(node.declarators()[0].scopedName()[:-1]))

                self._header.out(TMPL_proxy_h_operation, name=self.makeCxxName(name),
                                 capname=capname,
                                 paramspec=paramspec, retspec='void')

                self._impl.out(TMPL_proxy_c_operation_header,
                               name=self.makeCxxName(name),
                               fqname=self.makeCxxName(self.fqname),
                               capname=capname,
                               paramspec=paramspec,
                               retspec='void')
                self._impl.inc_indent()

                self._impl.out(xpcom_preinvoke)
                objname = self.getAppropriateObject(node)
                self._impl.out('_result = ' + objname +\
                               '->@capname@(@paramlist@);',
                               capname=capname, paramlist=callparams)

                self._impl.out(xpcom_postinvoke)

                self._impl.dec_indent()
                self._impl.out(TMPL_proxy_c_operation_footer)


    def visitOperation(self, node):
        name = node.identifier()
        fqname = idlutil.ccolonName(node.scopedName()[:-1])
        capname = string.upper(name[0]) + name[1:]


        paramspec = []
        callparams = []
        xpcom_postinvoke = []
        xpcom_preinvoke = []
        for p in node.parameters():
            corba_type = xpcorba_type.get_type(p.paramType())

            paramspec.append(corba_type.decl_corba(p.identifier(),
                                                   p.is_in(), p.is_out()))
            callparams.append(corba_type.call_xpcom_param(p.identifier(),
                                                          p.is_in(),
                                                          p.is_out()))

            corba_xpcom_preinvoke = corba_type.xpcom_preinvoke(p.identifier(),
                                                               p.is_in(),
                                                               p.is_out(),
                                                               node.oneway())
            corba_xpcom_postinvoke = corba_type.xpcom_postinvoke(p.identifier(),
                                                                 p.is_in(),
                                                                 p.is_out(),
                                                                 node.oneway())

            if corba_xpcom_preinvoke != '':
                xpcom_preinvoke.append(corba_xpcom_preinvoke)
            if corba_xpcom_postinvoke != '':
                xpcom_postinvoke.append(corba_xpcom_postinvoke)

        rettype = node.returnType()
        if not (isinstance(rettype, idltype.Base) and
                (rettype.kind() == idltype.tk_void)):
            corba_type = xpcorba_type.get_type(rettype)

            retspec = corba_type.decl_corba('', 0, 1)

            callparams.append(corba_type.call_xpcom_param('_retval', 0, 1))

            corba_xpcom_preinvoke = corba_type.xpcom_preinvoke('_retval', 0, 1, node.oneway())
            corba_xpcom_postinvoke = corba_type.xpcom_postinvoke('_retval',
                                                                 0, 1, node.oneway())

            if corba_xpcom_preinvoke != '':
                xpcom_preinvoke.append(corba_xpcom_preinvoke)
            if corba_xpcom_postinvoke != '':
                xpcom_postinvoke.append(corba_xpcom_postinvoke)
        else:
            retspec = 'void'

        self._header.out(TMPL_proxy_h_operation, name=self.makeCxxName(name),
                         capname=capname,
                         paramspec=string.join(paramspec, ', '),
                         retspec=retspec)

        self._impl.out(TMPL_proxy_c_operation_header,
                       name=self.makeCxxName(name),
                       fqname=self.makeCxxName(self.fqname), capname=capname,
                       paramspec=string.join(paramspec, ', '), retspec=retspec)
        self._impl.inc_indent()

        if retspec != 'void':
            self._impl.out(corba_type.type_corba() + ' _retval;')

        self._impl.out(string.join(xpcom_preinvoke, '\n'))
        objname = self.getAppropriateObject(node)
        self._impl.out('_result = ' + objname + '->@capname@(@paramlist@);',
                       capname=capname,
                       paramlist=string.join(callparams, ', '))

        self._impl.out(string.join(xpcom_postinvoke, '\n'))
        if retspec != 'void':
            self._impl.out('return _retval;')

        self._impl.dec_indent()
        self._impl.out(TMPL_proxy_c_operation_footer)
