/*	-*- C++ -*-
 * Copyright (C) 2001,  Christof Meerwald <cmeerw@web.de>
 * http://xpcom2corba.sunsite.dk/
 */

/*
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License
 * at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and
 * limitations under the License.
 *
 * The Original Code is the XPCOM-CORBA Bridge package.
 *
 * The Initial Developer of the Original Code is Christof Meerwald.
 * Portions created by Christof Meerwald are Copyright (C) 2001
 * Christof Meerwald.  All Rights Reserved.
 *
 * Contributor(s): Christof Meerwald <cmeerw@web.de> (original author)
 */

#include "xpcom_object.h"
#include "xpcom_proxy.h"
#include "corba2xpcom.h"
#include "nsIProgrammingLanguage.h"
#ifndef MOZILLA_INTERNAL_API
#include "nsCRTGlue.h"
#else
#include <nsCRT.h>
#endif

#include <ICORBAWrapper.h>


class CORBAWrapper_Impl
  : public ICORBAWrapper
{
 public:
  NS_DECL_ISUPPORTS
  NS_DECL_ICORBAWRAPPER


  CORBAWrapper_Impl(CORBA::Object_ptr corba_obj)
    : corba_obj_(corba_obj)
  { }

  virtual ~CORBAWrapper_Impl();


 private:
  CORBA::Object_var corba_obj_;
};


NS_IMPL_THREADSAFE_ISUPPORTS1(CORBAWrapper_Impl, ICORBAWrapper)

CORBAWrapper_Impl::~CORBAWrapper_Impl()
{
}

NS_IMETHODIMP CORBAWrapper_Impl::GetCorbaObj(CORBA::Object_ptr *aCorbaObj)
{
  *aCorbaObj = CORBA::Object::_duplicate(corba_obj_);

  XPCOM::IObject_var ip = XPCOM::IObject::_narrow(corba_obj_);
  if (!CORBA::is_nil(ip))
    ip->add_ref();

  return NS_OK;
}

NS_IMETHODIMP CORBAWrapper_Impl::GetObjid(char** id)
{
  try
  {
    XPCOM::IObject_var xpcom_obj_(XPCOM::IObject::_narrow(corba_obj_));
    if (CORBA::is_nil(xpcom_obj_))
      return NS_ERROR_FAILURE;
    char* corba_str = xpcom_obj_->objid();
    if (corba_str == NULL)
      return NS_ERROR_OUT_OF_MEMORY;
    ::corba2xpcom(corba_str, id);
    ::CORBA::string_free(corba_str);
    return NS_OK;
  }
  catch (...)
  {
    return NS_ERROR_FAILURE;
  }
}

NS_IMPL_THREADSAFE_ADDREF(CORBA_IObject__XPCOM_Proxy)
NS_IMPL_THREADSAFE_RELEASE(CORBA_IObject__XPCOM_Proxy)
NS_IMPL_QUERY_HEAD(CORBA_IObject__XPCOM_Proxy)
  if (aIID.Equals(NS_GET_IID(CORBA_IObject)))
  {
    foundInterface = NS_STATIC_CAST(nsISupports *, this);
  }
  else if (aIID.Equals(NS_GET_IID(ICORBAWrapper)))
  {
    foundInterface = new CORBAWrapper_Impl(CORBA::Object::_duplicate(corba_obj_));
  }
  else if (aIID.Equals(NS_GET_IID(nsIClassInfo)))
  {
    foundInterface = NS_STATIC_CAST(nsIClassInfo*, this);
  }
  else
  {
    PRInt32* rc = new PRInt32(1);
    // It will get incremented by itself, so start at zero...
    *rc = 0;
    foundInterface = XPCOM_Proxy_Factory::GetProxy(aIID, corba_obj_, rc);
  }

  nsresult status;
  if (foundInterface)
  {
    NS_ADDREF(foundInterface);
    status = NS_OK;
  }
  else
  {
    status = NS_NOINTERFACE;
  }
  *aInstancePtr = foundInterface;
  return status;
}

CORBA_IObject__XPCOM_Proxy::CORBA_IObject__XPCOM_Proxy(CORBA::Object_ptr corba_obj)
  : corba_obj_(CORBA::Object::_duplicate(corba_obj))
{
  NS_INIT_ISUPPORTS();
}

CORBA_IObject__XPCOM_Proxy::CORBA_IObject__XPCOM_Proxy(CORBA::Object_ptr corba_obj, PRInt32 *ref_count)
  : corba_obj_(CORBA::Object::_duplicate(corba_obj))
{
  NS_INIT_ISUPPORTS();
}

CORBA_IObject__XPCOM_Proxy::~CORBA_IObject__XPCOM_Proxy()
{
}

NS_IMETHODIMP
CORBA_IObject__XPCOM_Proxy::GetInterfaces(PRUint32 *aCount, nsIID*** aArray)
{
  // Getting the list not supported.
  *aCount = 0;
  *aArray = nsnull;
  return NS_OK;
}

NS_IMETHODIMP
CORBA_IObject__XPCOM_Proxy::GetHelperForLanguage
(
 PRUint32 language,
 nsISupports** aLangHelper
)
{
  *aLangHelper = nsnull;
  return NS_OK;
}

NS_IMETHODIMP
CORBA_IObject__XPCOM_Proxy::GetContractID(char** aContractID)
{
  *aContractID = nsnull;
  return NS_OK;
}

NS_IMETHODIMP
CORBA_IObject__XPCOM_Proxy::GetClassDescription(char** aClassDescription)
{
#ifndef MOZILLA_INTERNAL_API
  *aClassDescription =
    NS_strdup("[[CORBA_IObject__XPCOM_Proxy wrapped object]]");
#else
  *aClassDescription =
	  nsCRT::strdup("[[CORBA_IObject__XPCOM_Proxy wrapped object]]");
#endif
  return NS_OK;
}

NS_IMETHODIMP
CORBA_IObject__XPCOM_Proxy::GetClassID(nsCID** aClassID)
{
  *aClassID = nsnull;
  return NS_OK;
}

NS_IMETHODIMP
CORBA_IObject__XPCOM_Proxy::GetImplementationLanguage
(
 PRUint32* aImplementationLanguage
)
{
  *aImplementationLanguage = nsIProgrammingLanguage::CPLUSPLUS;
  return NS_OK;
}

NS_IMETHODIMP
CORBA_IObject__XPCOM_Proxy::GetFlags(PRUint32* aFlags)
{
  // All proxies are threadsafe at this end. They might not be at the other end
  // but we need to set this flag to avoid spurious warnings.
  *aFlags = nsIClassInfo::THREADSAFE;
  return NS_OK;
}

NS_IMETHODIMP
CORBA_IObject__XPCOM_Proxy::GetClassIDNoAlloc(nsCID *aClassIDNoAlloc)
{
  return NS_ERROR_NOT_AVAILABLE;
}

NS_IMPL_THREADSAFE_ADDREF(XPCOM_IObject__XPCOM_Proxy)
NS_IMPL_THREADSAFE_RELEASE(XPCOM_IObject__XPCOM_Proxy)
NS_IMPL_QUERY_HEAD(XPCOM_IObject__XPCOM_Proxy)
  if (aIID.Equals(NS_GET_IID(CORBA_IObject)))
  {
    foundInterface = NS_STATIC_CAST(nsISupports *, this);
  }
  else if (aIID.Equals(NS_GET_IID(ICORBAWrapper)))
  {
    foundInterface = new CORBAWrapper_Impl(CORBA::Object::_duplicate(corba_obj_));
  }
  else if (aIID.Equals(NS_GET_IID(nsIClassInfo)))
  {
    foundInterface = NS_STATIC_CAST(nsIClassInfo*, this);
  }
  else
  {
    XPCOM::IObject_var qi_ptr_;
    try
    {
      const char* iname = XPCOM_Proxy_Factory::GetInterfaceName(aIID);
      if (iname == NULL)
        return NS_ERROR_NO_INTERFACE;
      qi_ptr_ =
        corba_obj_->query_interface(iname);
    }
    catch (...)
    {
      return NS_ERROR_FAILURE;
    }
    if (CORBA::is_nil(qi_ptr_))
      return NS_ERROR_NO_INTERFACE;
    PRInt32* rc = new PRInt32(1);
    // It will get incremented by itself, so start at zero...
    *rc = 0;
    foundInterface = XPCOM_Proxy_Factory::GetProxy(aIID, qi_ptr_, rc);
  }

  nsresult status;
  if (foundInterface)
  {
    NS_ADDREF(foundInterface);
    status = NS_OK;
  }
  else
  {
    status = NS_NOINTERFACE;
  }
  *aInstancePtr = foundInterface;
  return status;
}

XPCOM_IObject__XPCOM_Proxy::XPCOM_IObject__XPCOM_Proxy(XPCOM::IObject_ptr corba_obj)
  : corba_obj_(XPCOM::IObject::_duplicate(corba_obj)),
    ref_count_(new PRInt32(1))
{
  *ref_count_ = 1;
  NS_INIT_ISUPPORTS();
}

XPCOM_IObject__XPCOM_Proxy::XPCOM_IObject__XPCOM_Proxy(XPCOM::IObject_ptr corba_obj, PRInt32 *ref_count)
  : corba_obj_(XPCOM::IObject::_duplicate(corba_obj)),
    ref_count_(ref_count)
{
  ++(*ref_count_);
  NS_INIT_ISUPPORTS();
}

XPCOM_IObject__XPCOM_Proxy::~XPCOM_IObject__XPCOM_Proxy()
{
  if (--(*ref_count_) <= 0)
  {
    delete ref_count_;

    try
    {
      corba_obj_->release_ref();
    }
    catch (...) {}
  }
}

NS_IMETHODIMP
XPCOM_IObject__XPCOM_Proxy::GetInterfaces(PRUint32 *aCount, nsIID*** aArray)
{
  // Getting the list not supported.
  *aCount = 0;
  *aArray = nsnull;
  return NS_OK;
}

NS_IMETHODIMP
XPCOM_IObject__XPCOM_Proxy::GetHelperForLanguage
(
 PRUint32 language,
 nsISupports** aLangHelper
)
{
  *aLangHelper = nsnull;
  return NS_OK;
}

NS_IMETHODIMP
XPCOM_IObject__XPCOM_Proxy::GetContractID(char** aContractID)
{
  *aContractID = nsnull;
  return NS_OK;
}

NS_IMETHODIMP
XPCOM_IObject__XPCOM_Proxy::GetClassDescription(char** aClassDescription)
{
#ifndef MOZILLA_INTERNAL_API
  *aClassDescription =
    NS_strdup("[[XPCOM_IObject__XPCOM_Proxy wrapped object]]");
#else
  *aClassDescription =
    nsCRT::strdup("[[XPCOM_IObject__XPCOM_Proxy wrapped object]]");
#endif
  return NS_OK;
}

NS_IMETHODIMP
XPCOM_IObject__XPCOM_Proxy::GetClassID(nsCID** aClassID)
{
  *aClassID = nsnull;
  return NS_OK;
}

NS_IMETHODIMP
XPCOM_IObject__XPCOM_Proxy::GetImplementationLanguage
(
 PRUint32* aImplementationLanguage
)
{
  *aImplementationLanguage = nsIProgrammingLanguage::CPLUSPLUS;
  return NS_OK;
}

NS_IMETHODIMP
XPCOM_IObject__XPCOM_Proxy::GetFlags(PRUint32* aFlags)
{
  // All proxies are threadsafe at this end. They might not be at the other end
  // but we need to set this flag to avoid spurious warnings.
  *aFlags = nsIClassInfo::THREADSAFE;
  return NS_OK;
}

NS_IMETHODIMP
XPCOM_IObject__XPCOM_Proxy::GetClassIDNoAlloc(nsCID *aClassIDNoAlloc)
{
  return NS_ERROR_NOT_AVAILABLE;
}
